#include "includes/easysock.h"
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <ctype.h>
#include <fcntl.h>

#ifndef _WIN32
	const int INVALID_SOCKET = -1;
#endif

/* Internal functions, to be used if platform is windows */
int sock_init(void) {
#ifdef _WIN32
	WSADATA wsa_data;
	return WSAStartup(MAKEWORD(1,1), &wsa_data);
#else
	return 0;
#endif
}
int sock_quit(void) {
#ifdef _WIN32
	return WSACleanup();
#else
	return 0;
#endif
}

/* Function to create a socket - Accepts IP version(4 or 6), protocol
type (PROTO_TCP or PROTO_UDP). */
SOCKET create_socket(int network, char transport) {
	sock_init();
	int domain;
	int type;

	if (network == 4) {
		domain = AF_INET;
	} else if (network == 6) {
		domain = AF_INET6;
	} else {
		return -1;
	}

	if (transport == ES_TCP) {
		type = SOCK_STREAM;
	} else if (transport == ES_UDP) {
		type = SOCK_DGRAM;
	} else {
		return -1;
	}

	int newSock = socket(domain,type,0);

	/* Set REUSEADDR flag, allowing program to be run twice */
	int set_opt = 1;
	setsockopt(newSock, SOL_SOCKET, SO_REUSEADDR, (char *)&set_opt, sizeof(set_opt));

	return newSock;
}


int create_addr(int network, const char* address, int port,struct sockaddr_storage* dest) {
	if (network == 4) {
		struct sockaddr_in listen_address;
		
		listen_address.sin_family = AF_INET;
		listen_address.sin_port = htons(port);
		inet_pton(AF_INET,address,&listen_address.sin_addr);
		memcpy((struct sockaddr *)dest,&listen_address,sizeof(listen_address));
		return 0;	

	} else if (network == 6) {
		struct sockaddr_in6 listen_ipv6;
		listen_ipv6.sin6_family = AF_INET6;
		listen_ipv6.sin6_port = htons(port);
		inet_pton(AF_INET6,address,&listen_ipv6.sin6_addr);
		memcpy((struct sockaddr_in6 *)dest,&listen_ipv6,sizeof(listen_ipv6));
		return 0;

	} else {
		return -202;
	}
		
}

SOCKET create_local (int network, char transport, const char* address, int port,struct sockaddr_storage* addr_struct) {
	int socket = create_socket(network,transport);
	if (socket < 0) {
		return (-1 * errno);
	}
	create_addr(network,address,port,addr_struct);
	int addrlen;
	if (network == 4) {
		addrlen = sizeof(struct sockaddr_in);
	} else if (network == 6) {
		addrlen = sizeof(struct sockaddr_in6);
	} else {
		return -202;
	}

	/* The value of addrlen should be the size of the 'sockaddr'. 
	This should be set to the size of 'sockaddr_in' for IPv4, and 'sockaddr_in6' for IPv6.
	See https://stackoverflow.com/questions/73707162/socket-bind-failed-with-invalid-argument-error-for-program-running-on-macos */

	int i = bind (socket,(struct sockaddr *)addr_struct,(socklen_t)addrlen);
	if (i < 0) {
		return (-1 * errno);
	}
	return socket;
}

SOCKET create_remote (int network,char transport, const char* address,int port,struct sockaddr_storage* remote_addr_struct) {

	struct addrinfo hints;  /* Used to tell getaddrinfo what kind of address we want */
	struct addrinfo* results; /* Used by getaddrinfo to store the addresses */


	if (check_ip_ver(address) < 0) { /* If the address is a domain name */
		int err_code;
		char* port_str = (char *)malloc(10 * sizeof(char));

		sprintf(port_str,"%d",port); /* getaddrinfo expects a string for its port */


		memset(&hints,'\0',sizeof(hints));
		hints.ai_socktype = char_to_socktype(transport);

		err_code = getaddrinfo(address,port_str,&hints,&results);
		if (err_code != 0) {
			return (-1 * err_code);
		}
		remote_addr_struct = (struct sockaddr_storage *)results->ai_addr;
		network = inet_to_int(results->ai_family);
	} else {
	        create_addr(network,address,port,remote_addr_struct);
	}

	int socket = create_socket(network,transport);
	if (socket < 0) {
                return (-1 * errno);
        }

       	int addrlen;
	if (network == 4) {
		addrlen = sizeof(struct sockaddr_in);
	} else if (network == 6) {
		addrlen = sizeof(struct sockaddr_in6);
	} else {
		return (-202);
	}

	/* The value of addrlen should be the size of the 'sockaddr'. 
	This should be set to the size of 'sockaddr_in' for IPv4, and 'sockaddr_in6' for IPv6.
	See https://stackoverflow.com/questions/73707162/socket-bind-failed-with-invalid-argument-error-for-program-running-on-macos */

        int i = connect(socket,remote_addr_struct,(socklen_t)addrlen);
	if (i < 0) {
		return (-1 * errno);
	}
        return socket;
}

/* sock_close - Closes the given socket */
int sock_close(SOCKET sock) {
	int status = 0;

	#ifdef _WIN32
		status = shutdown(sock, SD_BOTH);
		if (status == 0) { status = closesocket(sock); }
	#else
		status = shutdown(sock, SHUT_RDWR);
		if (status == 0) { status = close(sock); }
	#endif

	return status;
}


int check_ip_ver(const char* address) {
	char buffer[16]; /* 16 chars - 128 bits - is enough to hold an ipv6 address */
	if (inet_pton(AF_INET,address,buffer) == 1) {
		return 4;
	} else if (inet_pton(AF_INET6,address,buffer) == 1) {
		return 6;
	} else {
		return -1;
	}
}

int port_to_num(const char* port_str) {
	/* The largest possible port value is 65535: a 5 character string */
	if (strlen(port_str) > 5) {
		return -1;
	}
	
	for (int i = 0; i < strlen(port_str); i++) {
		if (isdigit(port_str[i]) == 0) { // Ensure that every character in port_str is a digit (isidigit() returns 0 if the parameter is not a digit)
			return -1;
		}
	}
	/* Convert the string to a base-10 integer */
	int port_num = (int)strtol(port_str, NULL, 10);
	if (port_num > 65535) {
		return -1;
	}
	if (port_num < 1024) {
		return -2;	
	}
	
	return port_num;
}

int int_to_inet(int network) {
	if (network == 4) {
		return AF_INET;
	} else if (network == 6) {
		return AF_INET6;
	} else {
		return -202;
	}
}

int inet_to_int(int af_type) {
	if (af_type == AF_INET) {
		return 4;
	} else if (af_type == AF_INET6) {
		return 6;
	} else {
		return -207;
	}
}

int char_to_socktype(char transport) {
	if (transport == ES_TCP) {
		return SOCK_STREAM;
	} else if (transport == ES_UDP) {
		return SOCK_DGRAM;
	} else {
		return -250;
	}
}
